package trapmon;

import java.awt.*;
import java.awt.event.*;
import java.awt.print.*;
import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.swing.text.html.*;


// ============================================================================
/**
 * Main frame window for SNMP Trap/Message Console.
 * Description and comments to follow.
 *
 * @author Alex Taylor, copyright (C) 2003
 *
 * @version 0.2
 */
public class JTrapFrame extends JFrame
{
    // Constants
    private static final int SNMP_MESSAGE_SOCKET = 161;
    private static final int SNMP_TRAP_SOCKET    = 162;

    private static final String about_message = "SNMP Message Console v0.2\n\nCopyright (C) 2003 Alex Taylor\n\n";

    // Threads
    private ListenThread trapThread;     // Thread to listen for SNMP traps
    private ListenThread messageThread;  // Thread to listen for other SNMP messages (n.y.impl.)

    // User interface components
    protected JMenuBar      menu_bar;
    protected JMenu         program_menu,
                            help_menu;
    protected JTabbedPane   notebook;
    protected JSplitPane    message_split,
                            trap_split;
    protected JMessageTable message_table;
    protected JTrapTable    trap_table;
    protected JEditorPane   message_console,
                            trap_console;
    protected JTextField    status; 
    
    protected StyleSheet    console_style;

    // Actions
    protected StartAction  start_listening;
    protected StopAction   stop_listening;
    protected ClearAction  clear;
    protected PrintAction  save;
    protected QuitAction   quit;
    protected AboutAction  show_about;


    // ------------------------------------------------------------ constructor
    /**
     * Create application UI and start listening for traps.
     *
     */
    public JTrapFrame( ) {
        buildInterface();
        addListeners();
        messageThread = new ListenThread( SNMP_MESSAGE_SOCKET );
        trapThread = new ListenThread( SNMP_TRAP_SOCKET );
        receiveTraps();
        receiveMessages();
    } // ----------------------------------------------------------------------


    // --------------------------------------------------------- buildInterface
    /*
     * Populate the user interface.
     *
     */
    private void buildInterface( ) {
        JScrollPane   scroll1,
                      scroll2,
                      scroll3,
                      scroll4;
        Container     content;
        HTMLEditorKit kit;
        
        // Initialize the window contents
        content = getContentPane();
        content.setLayout( new BorderLayout( 2, 5 ));

        // Create the various parts of the interface...
        createActions();
        buildMenuBar();

        console_style = new StyleSheet();
        console_style.addRule("BODY { font-family: sans-serif; }");
        console_style.addRule("H1,H2,H3 { font-weight: bold; }");
        console_style.addRule("H1 { font-size: large; padding: 0.1cm; margin-bottom: 0.4cm; color: #000060; background-color: #e0e8f0; }");
        console_style.addRule("H2 { font-size: large; padding: 0.1cm; margin-bottom: 0.4cm; color: #000060; background-color: #e0e8f0; }");
        console_style.addRule("H3 { margin-bottom: 0.05cm; color: #000060; }");
        console_style.addRule("TD { padding-left: 0.1cm; }");
        console_style.addRule("TD.sitem { padding-right: 0.3cm; }");
        console_style.addRule("TD.vitem { padding-right: 0.3cm; }");

        kit = new HTMLEditorKit();        
        kit.setStyleSheet( console_style );

        message_table = new JMessageTable();
        scroll1 = new JScrollPane( message_table );
        scroll1.setHorizontalScrollBarPolicy( JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
        scroll1.setVerticalScrollBarPolicy(   JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED   );
        scroll1.setMinimumSize( new Dimension( 0, 0 ));

        message_console = new JEditorPane();
        message_console.setEditorKit( kit );

        message_console.setEditable( false );
        scroll2 = new JScrollPane( message_console );
        scroll2.setHorizontalScrollBarPolicy( JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
        scroll2.setVerticalScrollBarPolicy(   JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED   );
        scroll2.setMinimumSize( new Dimension( 0, 0 ));

        message_split = new JSplitPane( JSplitPane.VERTICAL_SPLIT, true, scroll1, scroll2 );

        trap_table = new JTrapTable();
        scroll3 = new JScrollPane( trap_table );
        scroll3.setHorizontalScrollBarPolicy( JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
        scroll3.setVerticalScrollBarPolicy(   JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED   );
        scroll3.setMinimumSize( new Dimension( 0, 0 ));

        trap_console = new JEditorPane();
        trap_console.setEditorKit( kit );

        trap_console.setEditable( false );
        scroll4 = new JScrollPane( trap_console );
        scroll4.setHorizontalScrollBarPolicy( JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED );
        scroll4.setVerticalScrollBarPolicy(   JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED   );
        scroll4.setMinimumSize( new Dimension( 0, 0 ));

        trap_split = new JSplitPane( JSplitPane.VERTICAL_SPLIT, true, scroll3, scroll4 );

        notebook = new JTabbedPane( JTabbedPane.BOTTOM );
        notebook.addTab("Traps", trap_split );
        notebook.addTab("Other Messages", message_split );
        content.add("Center", notebook );

        status = new JTextField();
        status.setEditable( false );
        content.add("South", status );

        this.setBounds( 0, 26, 636, 450 );
        trap_split.setDividerLocation( 100 );
        message_split.setDividerLocation( 100 );

    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- buildMenuBar
    /*
     * Create the menubar.
     *
     */
    private void buildMenuBar( ) {
        JMenuItem item;

        menu_bar = new JMenuBar();
        program_menu = new JMenu("Program");
//        item = program_menu.add( start_listening );
//        item = program_menu.add( stop_listening );
//        program_menu.addSeparator();
        item = program_menu.add( save );
        item = program_menu.add( clear );
        program_menu.addSeparator();
        item = program_menu.add( quit );
        help_menu = new JMenu("Help");
        item = help_menu.add( show_about );

        menu_bar.add( program_menu );
        menu_bar.add( help_menu    );

        this.setJMenuBar( menu_bar );
    } // ----------------------------------------------------------------------


    // ---------------------------------------------------------- createActions
    /*
     * Create the Action objects.
     *
     */
    private void createActions( ) {
//        start_listening = new StartAction( this );
//        stop_listening  = new StopAction( this );
        save            = new PrintAction( this );
        clear           = new ClearAction( this );
        quit            = new QuitAction( this );
        show_about      = new AboutAction( this );
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- addListeners
    /*
     * Add the UI event listeners.
     *
     */
    private void addListeners( ) {
        ListSelectionModel mTableSelect,
                           tTableSelect;

        this.addWindowListener(
            new WindowAdapter( ) {
                public void windowClosing( WindowEvent e ) {
                    quitProgram();
                }
            }
        );

        mTableSelect = message_table.getSelectionModel();
        mTableSelect.addListSelectionListener( 
            new ListSelectionListener( ) {
                public void valueChanged( ListSelectionEvent e ) {
                    if ( ! e.getValueIsAdjusting() ) {
                        SnmpMessage m = message_table.getSelectedMessage();
                        if ( m == null ) return;
                        switch( m.type ) {
                            case SnmpMessage.ASN_SNMP_GETREQUEST: 
                            case SnmpMessage.ASN_SNMP_GETNEXTREQUEST: 
                            case SnmpMessage.ASN_SNMP_GETRESPONSE:
                            case SnmpMessage.ASN_SNMP_SETREQUEST:
                                showMessageDetails( m ); break;
                            default: break;    
                        }    
                    }
                }
            }
        );    
        
        tTableSelect = trap_table.getSelectionModel();
        tTableSelect.addListSelectionListener( 
            new ListSelectionListener( ) {
                public void valueChanged( ListSelectionEvent e ) {
                    if ( ! e.getValueIsAdjusting() ) {
                        SnmpMessage m = trap_table.getSelectedMessage();
                        if ( m == null ) return;
                        switch( m.type ) {
                            case SnmpMessage.ASN_SNMP_TRAP:
                                showTrapDetails( m ); break;
                            default: break;    
                        }    
                    }
                }
            }
        );    

    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- receiveTraps
    /**
     * Start listening for incoming traps.
     *
     */
    protected void receiveTraps( ) {
        status.setText("Listening for messages ...");
        trapThread.start();
//        start_listening.setEnabled( false );
//        stop_listening.setEnabled( true );
    } // ----------------------------------------------------------------------


    // ------------------------------------------------------------ rejectTraps
    /**
     * Stop listening for incoming traps.
     *
     */
    protected void rejectTraps( ) {
        status.setText("Not listening for messages");
        trapThread.end();
//        start_listening.setEnabled( true );
//        stop_listening.setEnabled( false );
    } // ----------------------------------------------------------------------


    // -------------------------------------------------------- receiveMessages
    /**
     * Start listening for incoming messages.
     *
     */
    protected void receiveMessages( ) {
        status.setText("Listening for messages ...");
        messageThread.start();
//        start_listening.setEnabled( false );
//        stop_listening.setEnabled( true );
    } // ----------------------------------------------------------------------


    // ------------------------------------------------------------ rejectTraps
    /**
     * Stop listening for incoming messages.
     *
     */
    protected void rejectMessages( ) {
        status.setText("Not listening for messages");
        messageThread.end();
//        start_listening.setEnabled( true );
//        stop_listening.setEnabled( false );
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- parseMessage
    /** 
     * Parse & interpret an incoming SNMP message.
     *
     */
    protected boolean parseMessage( byte[] message, InetAddress sender ) {
        StringBuffer dataBuf;
        EncodedElement contents,
                       version,
                       community,
                       pduData;
        SnmpMessage snmpMsg;               
        int offset;

        /* System.out.println( SMIHandler.getByteString( message )); */

        // Read the message contents
        contents = SMIHandler.decodeElement( message, 0 ); 

        // Trap should be SEQUENCE { version, community, data } -- verify
        if ( contents.identifier != SMIHandler.ASN_SEQUENCE ) return ( false );
        offset = 0;

        // Decode the version element
        version = SMIHandler.decodeElement( contents.data, offset );
        offset += (version.offsetinc + version.length);

        // TODO: if version != our designated version return false

        // Decode the community element
        community = SMIHandler.decodeElement( contents.data, offset );
        offset += (community.offsetinc + community.length);

        // TODO: if community not in our acceptance list return false

        // Decode the message data
        pduData = SMIHandler.decodeElement( contents.data, offset );
        snmpMsg = new SnmpMessage( version, community, pduData, sender );
        if ( ! snmpMsg.isValid ) return ( false );

        if ( snmpMsg.type == SnmpMessage.ASN_SNMP_TRAP )
            trap_table.addMessage( snmpMsg );
        else if (( snmpMsg.type == SnmpMessage.ASN_SNMP_GETREQUEST )     ||
                 ( snmpMsg.type == SnmpMessage.ASN_SNMP_GETNEXTREQUEST ) ||
                 ( snmpMsg.type == SnmpMessage.ASN_SNMP_GETRESPONSE )    ||
                 ( snmpMsg.type == SnmpMessage.ASN_SNMP_SETREQUEST )        )
            message_table.addMessage( snmpMsg );
        else return ( false );    

        return ( true );
    } // ----------------------------------------------------------------------


    // ------------------------------------------------------------------------
    /**
     *  Display the details of a trap on the trap console.
     *
     */
    protected void showTrapDetails( SnmpMessage snmpTrap ) {
        StringBuffer trap_content;
        int          i;

        double seconds = snmpTrap.timeTicks / 100;                     
        trap_content   = new StringBuffer();
        trap_content.append("<HTML><BODY>\n");
        trap_content.append("<TABLE width=100% border=0 cellspacing=0 cellpadding=0>");
        trap_content.append("<TR><TD><H1><b>SNMP TRAP</b></H1></TD>\n");
        trap_content.append("    <TD width=1% align=right><H2>" + snmpTrap.originator.getHostAddress() + "</H2></TD></TR>\n");
        trap_content.append("</TABLE></P>\n");
        trap_content.append("<TABLE border=0 cellspacing=0 cellpadding=0 class=\"summary\">\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Version:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + Integer.toString( snmpTrap.version ) + "</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Community:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + snmpTrap.community                   + "</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Agent:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + snmpTrap.agentAddress                + "</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Enterprise:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + snmpTrap.numericOID + "&nbsp; (" + snmpTrap.textualOID + ")</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Generic-trap:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + snmpTrap.genericTrapName + "&nbsp; (" + Integer.toString( snmpTrap.genericTrapNumber ) + ")</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Specific-trap:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + Integer.toString( snmpTrap.specificTrapNumber ) + "</TD></TR>\n");
        trap_content.append("<TR><TD class=\"sitem\"><i>Time-stamp:</i> &nbsp; &nbsp; </TD>\n");
        trap_content.append("    <TD class=\"svalue\">" + Double.toString( seconds )           + " s</TD></TR>\n");
        trap_content.append("</TABLE><P></DIV>\n\n");
        
        if ( ! snmpTrap.variables.isEmpty() ) {
            trap_content.append("<H3><b>Trap-Specific Variables:</b></H3>");
            trap_content.append("<DIV>");
            trap_content.append("<TABLE border=0 cellspacing=0 cellpadding=0 class=\"variables\">\n");
            for ( i = 0; i < snmpTrap.variables.size(); i++ ) {
                VariableBinding var = (VariableBinding) snmpTrap.variables.elementAt( i );
                trap_content.append("<TR><TD class=\"vitem\"><i>" + var.textualOID + ":</i> &nbsp; &nbsp; </TD>\n");
                trap_content.append("    <TD class=\"vvalue\">" + var.textValue + "</TD></TR>\n");
            }
            trap_content.append("</TABLE></DIV>\n\n");
        }    

        trap_content.append("</BODY></HTML>");
        trap_console.setText( trap_content.toString() );

    } // ----------------------------------------------------------------------


    // ------------------------------------------------------------------------
    /**
     *  Display the details of a message on the message console.
     *
     */
    protected void showMessageDetails( SnmpMessage snmpMsg ) {
        StringBuffer message_content;    
        int          i;
        
        message_content = new StringBuffer();
        message_content.append("<HTML><BODY>\n");
        message_content.append("<TABLE width=100% border=0 cellspacing=0 cellpadding=0>");
        message_content.append("<TR><TD><H1><b>SNMP MESSAGE</b></H1></TD>\n");
        message_content.append("    <TD width=1% align=right><H2>" + snmpMsg.originator.getHostAddress() + "</H2></TD></TR>\n");
        message_content.append("</TABLE></P>\n");
        message_content.append("<TABLE border=0 cellspacing=0 cellpadding=0 class=\"summary\">");
        message_content.append("<TR><TD class=\"sitem\"><i>Version:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + Integer.toString( snmpMsg.version )     + "</TD></TR>\n");
        message_content.append("<TR><TD class=\"sitem\"><i>Community:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + snmpMsg.community                       + "</TD></TR>\n");
        message_content.append("<TR><TD class=\"sitem\"><i>Message type:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + snmpMsg.messageName                     + "</TD></TR>\n");

        message_content.append("<TR><TD class=\"sitem\"><i>Request ID:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + Long.toString( snmpMsg.requestID )      + "</TD></TR>\n");

        message_content.append("<TR><TD class=\"sitem\"><i>Error status:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + snmpMsg.errorStatusName + "&nbsp; ("
                                                           + Integer.toString( snmpMsg.errorStatus ) + ")</TD></TR>\n");
        message_content.append("<TR><TD class=\"sitem\"><i>Error index:</i> &nbsp; &nbsp; <TD>\n");
        message_content.append("    <TD class=\"svalue\">" + Integer.toString( snmpMsg.errorIndex )  + "</TD></TR>\n");
        message_content.append("</TABLE><P></DIV>\n\n");
        
        if ( ! snmpMsg.variables.isEmpty() ) {
            message_content.append("<H3><b>Message-Specific Variables:</b></H3>\n");
            message_content.append("<DIV><TABLE border=0 cellspacing=0 cellpadding=0 class=\"variables\">\n");
            for ( i = 0; i < snmpMsg.variables.size(); i++ ) {
                VariableBinding var = (VariableBinding) snmpMsg.variables.elementAt( i );
                message_content.append("<TR><TD class=\"vitem\"><i>" + var.textualOID + ":</i> &nbsp; &nbsp; </TD>\n");
                message_content.append("    <TD class=\"vvalue\">" + var.textValue + "</TD></TR>\n");
            }
            message_content.append("</TABLE></DIV>\n\n");
        }    

        message_content.append("</BODY></HTML>");
        message_console.setText( message_content.toString() );
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- clearConsole
    /**
     * Clear the message list, details panel, and status field.
     *
     */
    protected void clearConsole( ) {
        trap_table.clearTable();
        trap_console.setText("");
        message_table.clearTable();
        message_console.setText("");
        status.setText("");
    } // ----------------------------------------------------------------------


    // ---------------------------------------------------------- printSelected
      /*
       * Print the currently-active message.
       *
       */
    protected void printSelected( ) {
        JSplitPane page = (JSplitPane) notebook.getSelectedComponent();
        String text;
        PrintJob job;
        Graphics g;
        
        Toolkit tk = Toolkit.getDefaultToolkit();
        if ( page == trap_split ) {
            SnmpMessage m = trap_table.getSelectedMessage();
            if ( m != null ) {
                job = tk.getPrintJob( this, "Print trap details", null );
                if ( job != null ) {
                    g = job.getGraphics();
                    g.translate( 15, 15 );
                    trap_console.paint( g );
                    g.dispose();
                    job.end();
                }    
            } else
                JOptionPane.showMessageDialog( this, "No trap is currently selected.",
                                               "Error", JOptionPane.ERROR_MESSAGE     );
        } else if ( page == message_split ) {
            SnmpMessage m = message_table.getSelectedMessage();
            if ( m != null ) {
                job = tk.getPrintJob( this, "Print message details", null );
                if ( job != null ) {
                    g = job.getGraphics();
                    g.translate( 15, 15 );
                    trap_console.paint( g );
                    g.dispose();
                    job.end();
                }    
//                text = message_console.getText();
//                System.out.println( text );
            } else
                JOptionPane.showMessageDialog( this, "No message is currently selected.",
                                               "Error", JOptionPane.ERROR_MESSAGE        );
        }
        
    } // ----------------------------------------------------------------------


    // ------------------------------------------------------------ quitProgram
      /*
       * Perform necessary cleanup and exit the program.
       *
       */
    protected void quitProgram( ) {
        rejectTraps();
        dispose();
        System.exit( 0 );
    } // ----------------------------------------------------------------------




    /* MEMBER CLASS DEFINITIONS */

    // ========================================================================
    class StartAction extends AbstractAction
    {
        JTrapFrame app;
        public StartAction( JTrapFrame parent ) {
            super("Start listening");
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            app.receiveTraps();
            app.receiveMessages();
        }
    } // ======================================================================



    // ========================================================================
    class StopAction extends AbstractAction
    {
        JTrapFrame app;
        public StopAction( JTrapFrame parent ) {
            super("Stop listening");
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            app.rejectTraps();
            app.rejectMessages();
        }
    } // ======================================================================



    // ========================================================================
    class ClearAction extends AbstractAction
    {
        JTrapFrame app;
        public ClearAction( JTrapFrame parent ) {
            super("Clear list");
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            app.clearConsole();
        }
    } // ======================================================================


    // ========================================================================
    class PrintAction extends AbstractAction
    {
        JTrapFrame app;
        public PrintAction( JTrapFrame parent ) {
            super("Print selected...");
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            app.printSelected();
        }
    } // ======================================================================


    // ========================================================================
    class QuitAction extends AbstractAction
    {
        JTrapFrame app;
        public QuitAction( JTrapFrame parent ) {
            super("Quit");
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            app.quitProgram();
        }
    } // ======================================================================



    // ========================================================================
    class AboutAction extends AbstractAction
    {
        JTrapFrame app;
        public AboutAction( JTrapFrame parent ) {
            super("Product information", null );
            app = parent;
        }
        public void actionPerformed( ActionEvent e ) {
            JOptionPane.showMessageDialog( app, app.about_message, "Product Information",
                                           JOptionPane.INFORMATION_MESSAGE    );
        }
    } // ======================================================================



    // ========================================================================
    class ListenThread extends Thread
    {
        private volatile boolean listening;
        private int              socket;

        // --------------------------------------------------------------------
        ListenThread( int listenSocket ) {
            socket = listenSocket;
            listening = false;
        } // ------------------------------------------------------------------


        // --------------------------------------------------------------------
        public synchronized void run( ) {
            DatagramSocket listener;
            DatagramPacket incoming;
            InetAddress    source;
            byte    trapbuf[];
            boolean error,
                    processed;

            error = false;
            try {
                listener = new DatagramSocket( socket );
            } catch ( SocketException se ) {
                status.setText("Socket Exception" + se.getMessage() );
                listener = null;
                error = true;
            }

            if ( ! error ) {
                listening = true;
                trapbuf = new byte[ 484 ];
                incoming = new DatagramPacket( trapbuf, trapbuf.length );
                while ( listening ) {
                    try {
                        listener.receive( incoming );
                        source  = incoming.getAddress();
                        trapbuf = incoming.getData();
                        processed = parseMessage( trapbuf, source );
                        if ( ! processed ) 
                            status.setText("Rejected message");
                    } catch ( IOException ioe ) {
                        status.setText("I/O Exception:" + ioe.getMessage() );
                    }
                } 
                listener.close();
            }

        } // ------------------------------------------------------------------


        // --------------------------------------------------------------------
        public void end( ) {
            listening = false;
            this.interrupt();
        } // ------------------------------------------------------------------

    } // ======================================================================


} // ==========================================================================

