package trapmon;

import java.util.*;

/**
 * A collection of methods for interpreting ASN.1 elements used by 
 * the SMI specification.                                                   <P>
 *
 * This class implements the Structure of Management Information (SMI) 
 * specification of RFC 1155 (full title: "Structure and Identification of
 * Management Information for TCP/IP based Internets").                     <P>
 *
 * SMI is a specific implementation of the ASN.1 Abstract Syntax Notation 
 * (ISO 8824) utilizing the Basic Encoding Rules (ISO 8825); or rather, a
 * subset of those rules.  It forms the basis of SNMP message encoding.
 *
 * @author Alex Taylor, copyright (C) 2003
 */
public class SMIHandler extends Object
{

    // Common ASN.1/SMI object identifier prefixes

                /** 
                 * OID for "internet directory".                            <P>
                 * OBJECT IDENTIFIER ::= { internet directory(1) }           
                 */
    public static final String SMI_OID_DIRECTORY    = new String("1.3.6.1.1");
    
                /** 
                 * OID for "internet mgmt".                                 <P>
                 * OBJECT IDENTIFIER ::= { internet mgmt(2)      }           
                 */
    public static final String SMI_OID_MGMT         = new String("1.3.6.1.2");
    
                /** 
                 * OID for "internet experimental".                         <P>
                 * OBJECT IDENTIFIER ::= { internet experimental(3) }        
                 */
    public static final String SMI_OID_EXPERIMENTAL = new String("1.3.6.1.3");
    
                /** 
                 * OID for "internet private".                              <P>
                 * OBJECT IDENTIFIER ::= { internet private(4) }             
                 */
    public static final String SMI_OID_PRIVATE      = new String("1.3.6.1.4");
    
                /** 
                 * OID for "internet private enterprises".                  <P>
                 * OBJECT IDENTIFIER ::= { internet private enterprises(1) } 
                 */
    public static final String SMI_OID_ENTERPRISES  = new String("1.3.6.1.4.1");


    // Some ASN.1 element types (includes all bits of ID byte, for simplicity)

    // Universal ASN.1 types    
                /** 
                 * ASN.1 BOOLEAN type [UNIVERSAL 1].  Always uses primitive 
                 * encoding; binary value is '00 0 00001'.
                 */
    public static final int ASN_BOOLEAN                                  = 0x01; 
    
                /** 
                 * ASN.1 INTEGER type [UNIVERSAL 2].  Always uses primitive 
                 * encoding; binary value is '00 0 00010'.
                 */
    public static final int ASN_INTEGER                                  = 0x02; 
    
                /** 
                 * ASN.1 BIT STRING type [UNIVERSAL 3], primitive form. 
                 * Binary value is '00 0 00011'.
                 */
    public static final int ASN_PBITSTRING                               = 0x03; 

                /** 
                 * ASN.1 BIT STRING type [UNIVERSAL 3], constructed form. 
                 * Binary value is '00 1 00011'.
                 */
    public static final int ASN_CBITSTRING                               = 0x23; 
    
                /** 
                 * ASN.1 OCTET STRING type [UNIVERSAL 4], primitive form. 
                 * Binary value is '00 0 00100'.
                 */
    public static final int ASN_POCTETSTRING                             = 0x04; 

                /** 
                 * ASN.1 OCTET STRING type [UNIVERSAL 4], constructed form. 
                 * Binary value is '00 1 00100'.
                 */
    public static final int ASN_COCTETSTRING                             = 0x24; 
    
                /** 
                 * ASN.1 NULL type [UNIVERSAL 5].  Always uses primitive 
                 * encoding; binary value is '00 0 00101'.
                 */
    public static final int ASN_NULL                                     = 0x05; 
    
                /** 
                 * ASN.1 OBJECT IDENTIFIER type [UNIVERSAL 6].  Always uses 
                 * primitive encoding; binary value is '00 0 00110'.
                 */
    public static final int ASN_OBJECTIDENTIFIER                         = 0x06; 
    
                /** 
                 * ASN.1 REAL type [UNIVERSAL 9].  Always uses primitive 
                 * encoding; binary value is '00 0 01001'.
                 */
    public static final int ASN_REAL                                     = 0x09; 
    
                /** 
                 * ASN.1 SEQUENCE type [UNIVERSAL 16].  Always uses 
                 * constructed encoding; binary value is '00 1 10000'.
                 */
    public static final int ASN_SEQUENCE                                 = 0x30; 
    
                /** 
                 * ASN.1 SET type [UNIVERSAL 17].  Always uses constructed
                 * encoding; binary value is '00 1 10001'.
                 */
    public static final int ASN_SET                                      = 0x31; 


    // SMI-specific types
                /** 
                 * SMI IpAddress type [APPLICATION 0]. 
                 */
    public static final int ASN_SMI_IPADDRESS                            = 0x40; 
    
                /** 
                 * SMI Counter type [APPLICATION 1].   
                 */
    public static final int ASN_SMI_COUNTER                              = 0x41; 

                /** 
                 * SMI Gauge type [APPLICATION 2].     
                 */
    public static final int ASN_SMI_GAUGE                                = 0x42; 
    
                /** 
                 * SMI TimeTicks type [APPLICATION 3]. 
                 */
    public static final int ASN_SMI_TIMETICKS                            = 0x43; 
    
                /** 
                 * SMI Opaque type [APPLICATION 4].    
                 */
    public static final int ASN_SMI_OPAQUE                               = 0x44; 


    // Some helpful constants for bit manipulation
                /** Bit 1 of a byte ('00000001').  Used in bit manipulation. */
    public static final int BIT1 = 0x01;
                /** Bit 2 of a byte ('00000010').  Used in bit manipulation. */
    public static final int BIT2 = 0x02;
                /** Bit 3 of a byte ('00000100').  Used in bit manipulation. */
    public static final int BIT3 = 0x04;
                /** Bit 4 of a byte ('00001000').  Used in bit manipulation. */
    public static final int BIT4 = 0x08;
                /** Bit 5 of a byte ('00010000').  Used in bit manipulation. */
    public static final int BIT5 = 0x10;
                /** Bit 6 of a byte ('00100000').  Used in bit manipulation. */
    public static final int BIT6 = 0x20;
                /** Bit 7 of a byte ('01000000').  Used in bit manipulation. */
    public static final int BIT7 = 0x40;
                /** Bit 8 of a byte ('10000000').  Used in bit manipulation. */
    public static final int BIT8 = 0x80;

    
    // ---------------------------------------------------------- decodeElement
    /**
     * Decodes the contents of a single ASN.1 syntactic element.  An SNMP   
     * message is an example of such an element.  Note that the element may
     * be (and frequently is) a compound data type such as SEQUENCE, in 
     * which case it encapsulates additional child elements.                <P>
     *
     * <TABLE border=1 cellspacing=0 cellpadding=10 bgcolor=#f0fff0><TR><TD>
     * <H3>Decoding of ASN.1 Elements</H3>
     *
     * What follows is a brief technical description of how ASN.1 element 
     * parsing works.                                                       <P>
     *
     * An ASN.1 element is encoded as follows:                             <UL>
     * <li>  1 or more identifier byte(s), followed by
     * <li>  1 or more length byte(s), followed by
     * <li>  0 or more content byte(s)                                    </UL>   
     *                                                                      <P>
     * <H4>Identifier Bytes</H4> 
     *
     * The identifier byte(s) indicate the class, encoding form, and data 
     * type (or 'tag') of the element.                                      <P>
     *
     * Bits 8 and 7 of the first byte indicate the element class:          <UL>
     * <li>  00 = universal
     * <li>  01 = application
     * <li>  10 = context-specific
     * <li>  11 = private                                                 </UL>
     *                                                                      <P>
     * Bit 6 represents the element encoding form:                         <UL>
     * <li>  0 = primitive
     * <li>  1 = constructed                                              </UL>
     *                                                                      <P>
     * Bits 5 to 1 form a binary number called the identifier tag.          <P>
     *
     * If the identifier tag is in the range of 0-30, then there are no 
     * further identifier bytes.  The tag represents the element data
     * type number.  The next byte will be the first length byte.           <P>
     *
     * If the identifier tag is 31 (all 1s), then the data type number 
     * is encoded in one or more subsequent identifier bytes, as follows:  <UL>
     * <li>  Bit 8 of each byte is 1 unless it is the last identifier 
     *       byte, in which case it is 0.
     * <li>  Bits 7 to 1 of each byte, including the last, are 
     *       concatenated together to form a binary integer which 
     *       represents the element data type number.                     </UL>
     *                                                                      <P>
     * Some common tag numbers in the UNIVERSAL class are listed below.  
     * This is not a complete list; refer to the ASN.1 standards for more 
     * information.  (P) indicates that primitive encoding is used for this
     * data type, (C) indicates constructed encoding, and (*) indicates 
     * that either form may be used.  Since all these values are below 31, 
     * the tag number will be part of the first identifier byte; the byte 
     * value column indicates the total value of this byte when the tag 
     * number is combined with the leading bits.
     * 
     *   <PRE>
     *   <B>Tag                    Byte Value                Type</B>
     *   [UNIVERSAL 1] (P)      0x01                      BOOLEAN
     *   [UNIVERSAL 2] (P)      0x02                      INTEGER
     *   [UNIVERSAL 3] (*)      0x03 (P) or 0x23 (C)      BIT STRING
     *   [UNIVERSAL 4] (*)      0x04 (P) or 0x24 (C)      OCTET STRING
     *   [UNIVERSAL 5] (P)      0x05                      NULL      
     *   [UNIVERSAL 6] (P)      0x06                      OBJECT IDENTIFIER
     *   [UNIVERSAL 9] (P)      0x09                      REAL             
     *   [UNIVERSAL 10] (P)     0x0A                      ENUMERATED
     *   [UNIVERSAL 16] (C)     0x30                      SEQUENCE or SEQUENCE OF
     *   [UNIVERSAL 17] (C)     0x31                      SET or SET OF
     *   </PRE>
     *                                                                      <P>
     * <H4>Length Bytes</H4>
     * 
     * The length bytes are used to indicate the number of content bytes in 
     * the element.                                                         <P>
     *
     * SNMP uses only the definite-length encoding format of ASN.1, of 
     * which there are two forms: short and long.  Bit 8 of the (first)
     * length byte indicates which form each element uses:                 <UL>
     * <li>  0 = short form
     * <li>  1 = long form                                                </UL>
     *                                                                      <P>
     * Short form indicates that there are no additional length bytes. 
     * Bits 7-1 thus represent a binary integer which indicates the number
     * of content bytes in the element.                                     <P>
     *
     * Long form indicates that one or more additional length bytes follow.
     * Bits 7-1 represent a binary integer which indicates the number of 
     * <em>additional</em> length bytes following the first one.  (If this
     * number were 0, it would indicate ASN.1 'indefinite form', in which 
     * case the contents are delimited by an 'end of content' marker; the 
     * SNMP standard, however, does not allow indefinite form.)  The bits
     * from all of the subsequent length bytes are concatenated together to 
     * form a single binary integer which indicates the number of content 
     * bytes in the element.                                                <P>
     *
     * <H4>Content Bytes</H4>
     * 
     * The content bytes contain the actual element data.  These bytes 
     * begin immediately following the last length byte.  The last content 
     * byte is also the last byte of the element as a whole.                <P>
     *
     * The meaning of whatever data is contained in the content bytes 
     * depends entirely on the data type of the element (as indicated by 
     * the identifier byte(s)), and may be application-specific.            <P>
     *
     * Note that if the element represents a compound type, such as SEQUENCE
     * (similar to a data structure), then the content bytes will themselves 
     * represent ASN.1 elements; the first content byte would thus represent
     * the first identifier byte of the first child element.  Child elements
     * should be decoded in turn, as needed.                                <P>
     *                                                                  <HR><P>
     * <H4>Example</H4>                                                     <P>
     *
     * Let's say we receive an SNMP message which looks like:
     *                                                                    <PRE>
     *   30 82 01 03 02 01 00 04 06 70 75 82 6C 69 63 ... (etc.)      </PRE><P>
     *
     * Looking at the first (identifier) byte, 0x30 = 00110000 translates to:
     *                                                                    <PRE>
     *   Class = UNIVERSAL   (bit 8 = 0, bit 7 = 0)
     *   Type  = CONSTRUCTED (bit 6 = 1)
     *   Tag   = 0x10 or 16  (bits 5 through 1 = 10000)                  </PRE>
     *                                                                      <P>
     * [UNIVERSAL 16] is the SEQUENCE type.                                 <P>
     *
     * The next byte is the first length byte.  We have 0x82 = 10000010.
     * Bit 8 = 1, therefore the length bytes use the long form.  Bits 7
     * through 1 represent the number of additional length bytes, which
     * in this case equals 2.                                               <P>
     *
     * Consequently, the bits of the next two bytes are concatenated 
     * together to form a new number which represents the total length of 
     * the contents.  0x01 = 00000001 and 0x03 = 00000011 becomes 100000011 
     * (we can ignore the leading zeroes), which is 259.  Therefore, there 
     * are 259 content bytes in this element, starting with the next byte
     * (0x02).                                                              <P>
     *  
     * The <i>decodeElement()</i> method returns with this information.  It 
     * is up to the application to deal with it as necessary.  In this 
     * particular case, the element in question is a SEQUENCE, meaning that 
     * the content bytes themselves represent additional ASN.1 elements.  
     * Presumably the application will recognize this, and call the 
     * <i>decodeElement()</i> method again on the content bytes.  If so, the 
     * following information will be decoded:                             <PRE>
     *   02 01 00                   Represents an INTEGER with value 0
     *   04 06 70 75 62 6C 69 63    Represents an OCTET STRING with value 70-75-62-6C-69-63
     *   (etc.)                                                          </PRE>
     *                                                                      <P>
     * According to the SNMP message definition (see RFC 1157), we should 
     * have SEQUENCE{ version, community, data } where 'version' is an 
     * INTEGER value representing the version number minus 1, 'community' 
     * is an OCTET STRING of ASCII characters, and 'data' varies depending 
     * on the specific message type.  Therefore, what we have decoded thus 
     * far is:                                                            <PRE>
     *   SEQUENCE{ version = 1, community = 'public', data = (still to be decoded) }
     *                                                                </PRE><P>
     *                                                    </TD></TR></TABLE><P>
     *
     * @param message An array of bytes containing the element to be 
     *   parsed.  This may be a raw SNMP message packet, or a subset; 
     *   however, it must represent a complete ASN.1 element.               <P>
     *
     * @param offset The offset into the array at which to begin decoding.
     *   This allows child elements to be parsed from within a larger byte
     *   array, without having to strip the leading bytes off.              <P>
     *
     * @return An <i>EncodedElement</i> object which represents the decoded 
     *   ASN.1 element.                                                     <P>
     *
     */
    public static EncodedElement decodeElement( byte[] message, int offset ) {
        EncodedElement element;
        StringBuffer bitBuffer;
        int i, j,
            ibyte,
            mask,
            identifier,
            length;
        byte data[];

        element = new EncodedElement();

        /*
         * A lot of this number parsing is made a great deal harder than it 
         * should have to be, thanks to Java's very annoying treatment of
         * binary numbers.
         */

        // Read the identifier byte(s)
        if (( ibyte = (int)message[offset++] ) < 0 ) ibyte += 256; 
        element.offsetinc = 1;
        mask = ( BIT1 | BIT2 | BIT3 | BIT4 | BIT5 );

        // If bits 1 through 5 are all 1 (value 31), then...
        if (( ibyte & mask ) == mask ) {

            /*
             * What we're doing here is reading n bytes in sequence, then
             * concatenating all of their bits together to form one long 
             * binary number.  We use a StringBuffer to manage this.
             */
            bitBuffer = new StringBuffer();
            do {        // We will keep reading bytes until bit 8 is 0.
                StringBuffer thisByte;

                element.offsetinc++;
                if (( ibyte = (int)message[offset++] ) < 0 ) ibyte += 256; 

                // Concatenate each byte (minus bit 8) onto our tag buffer.
                thisByte = new StringBuffer( Integer.toBinaryString( ibyte & ~BIT8 ));

                /*
                 * Pad the byte to 8 bits with leading 0's; the 
                 * Integer.toBinaryString() method seems to be unable to do 
                 * this for us.
                 */
                for ( j = thisByte.length(); j < 8; j++ ) {
                    thisByte.insert( 0, '0');
                }
                bitBuffer.append( thisByte );

            } while (( ibyte & BIT8 ) != BIT8 );
            
            // Convert our buffer into an binary integer value.
            identifier = Integer.parseInt( bitBuffer.toString(), 2 );

            /*
             * Convert to an unsigned number if necessary.  Doesn't Java
             * provide an easier way of doing this?
             */
            if ( identifier < 0 )
                identifier += (int)Math.pow( 2, bitBuffer.toString().length() - 1 );

        } // Otherwise, there is only 1 identifier byte; save it directly.
        else identifier = ibyte;

        element.identifier = identifier;

        // Read the length byte(s) to determine size of element data
        if (( ibyte = (int)message[offset++] ) < 0 ) ibyte += 256; 
        element.offsetinc++;
        if (( ibyte & BIT8 ) == BIT8 ) {        

            /*
             * If bit 8 == 1, then bits 1 through 7 tell us how many bytes
             * we have to read in order to get the length value.
             */
            int numBytes = ibyte & ~BIT8;

            /*
             * So now we do that whole concatenation of n bytes again...
             */ 
            bitBuffer = new StringBuffer();
            if ( numBytes < 0 ) numBytes += 64; 
            element.offsetinc += numBytes;

            for ( i = 0; i < numBytes; i++ ) {
                StringBuffer thisByte;
                if (( ibyte = (int)message[offset++] ) < 0 ) ibyte += 256; 
                thisByte = new StringBuffer( Integer.toBinaryString( ibyte ));
                for ( j = thisByte.length(); j < 8; j++ ) {
                    thisByte.insert( 0, '0');
                }
                bitBuffer.append( thisByte.toString() );
            }
            length = Integer.parseInt( bitBuffer.toString(), 2 );
            if ( length < 0 )          // convert from signed to unsigned
                length += (int)Math.pow( 2, bitBuffer.toString().length() - 1 );

        } else {
            // else length = bits 7-1
            length = ibyte;
        }
        element.length = length;

        // now read the content bytes
        data = new byte[ length ];
        for ( i = 0; i < length; i++ ) {
            data[ i ] = message[ i + offset ];
        } 
        element.data = data;

        return element;
    } // ----------------------------------------------------------------------


    // --------------------------------------------------------------- parseOID
    /**
     * Parses an element of type OBJECT IDENTIFIER.                         <P>
     *
     * An ASN.1 OBJECT IDENTIFIER is a sequence of dot-separated numbers, 
     * or subidentifiers, which represent a 'tree' of possible values. These 
     * can be used to identify companies, products, and product-specific 
     * variables.                                                           <P>
     * 
     * All SMI companies and variables fall under the 'internet' OID, which 
     * is:                                                         <BLOCKQUOTE>
     *     1.3.6.1                                                </BLOCKQUOTE>
     * For instance, all companies are identified by OIDs under the 
     * 'enterprises' tree, which is {internet private(4) enterprises(1) }, 
     * or                                                          <BLOCKQUOTE>
     *     1.3.6.1.4.1                                            </BLOCKQUOTE>
     *
     * This method doesn't check for value overruns, so a malformed value 
     * could cause an exception.                                            <P>
     *
     * @param bits The array of bytes to be parsed.                         <P>
     *
     * @return A <i>String</i> containing the dot-separated object 
     *   identifier sequence.                                               <P>
     *
     */
    public static String parseOID( byte[] bits )
    {
        StringBuffer bitBuffer,
                     oidBuffer;
        int i, 
            element,
            ibyte;
        long value,
             first,
             second;

        oidBuffer = new StringBuffer();
        bitBuffer = new StringBuffer();
        element = 1;
        for ( i = 0; i < bits.length; i++ ) {
            if (( ibyte = (int)bits[i] ) < 0 ) ibyte += 256; 
            bitBuffer.append( Integer.toBinaryString( ibyte & ~BIT8 ));
            if (( ibyte & BIT8 ) == 0 ) {
                value = Long.parseLong( bitBuffer.toString(), 2 );
                if ( value < 0 )
                    value += (long)Math.pow( 2, bitBuffer.toString().length() - 1 );
                if ( element == 1 ) {
                    // ASN.1 packs the first two subidentifiers into one
                    first = value / 40;
                    second = value % 40;
                    oidBuffer.append( Long.toString( first ) + ".");
                    oidBuffer.append( Long.toString( second ) + ".");
                }
                else oidBuffer.append( Long.toString( value ) + ".");
                bitBuffer = new StringBuffer();
                element++;
            }
        }

        // Trim off any trailing "."
        if ( oidBuffer.toString().endsWith("."))
            oidBuffer.deleteCharAt( oidBuffer.length()-1 );

        return ( oidBuffer.toString() );
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- translateOID
    /**
     * Attempts to translate a numeric OBJECT IDENTIFIER string into its
     * human-readable name.  (Should be part of a full MIB handler in future.)
     *
     * @param oid A <i>String</i> containing the dot-separated object
     *   identifier sequence.                                               <P>
     *
     * @return A <i>String</i> containing the human-readable object 
     *   identifier.                                                        <P>
     *
     */
    public static String translateOID( String oid )
    {
        StringBuffer oidBuffer = new StringBuffer( oid );
        
        // Try to translate the OID into a 'known' prefix...
        if ( oid.startsWith( SMI_OID_MGMT )) {
            oidBuffer.replace( 0, SMI_OID_MGMT.length(), "mgmt");
        }
        else if ( oidBuffer.toString().startsWith( SMI_OID_DIRECTORY )) {
            oidBuffer.replace( 0, SMI_OID_DIRECTORY.length(), "directory");
        }
        else if ( oidBuffer.toString().startsWith( SMI_OID_EXPERIMENTAL )) {
            oidBuffer.replace( 0, SMI_OID_EXPERIMENTAL.length(), "experimental");
        }
        else if ( oidBuffer.toString().startsWith( SMI_OID_ENTERPRISES )) {
            oidBuffer.replace( 0, SMI_OID_ENTERPRISES.length(), "enterprises");
        }
        else if ( oidBuffer.toString().startsWith( SMI_OID_PRIVATE )) {
            oidBuffer.replace( 0, SMI_OID_PRIVATE.length(), "private");
        }
        
        return ( oidBuffer.toString() );
    } // ----------------------------------------------------------------------


    // ---------------------------------------------------- parseNetworkAddress
    /**
     * Parses an element of the SMI type NetworkAddress.
     *
     * @param data The array of bytes to be parsed.                         <P>
     *
     * @return A <i>String</i> containing the network address.              <P>
     *
     */
    public static String parseNetworkAddress( byte[] data )
    {
        int i, ibyte;
        StringBuffer addrBuf;

        if ( data.length != 4 )
            return ("invalid network address");
        else {
            addrBuf = new StringBuffer();
            for ( i = 0; i < 4; i++ ) {
                if (( ibyte = (int)data[i] ) < 0 ) ibyte += 256; 
                addrBuf.append( Integer.toString( ibyte ));
                if ( i < 3 ) addrBuf.append(".");
            } 
            return ( addrBuf.toString() );
        }      
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- parseInteger
    /**
     * Translates an array of bytes into a single signed (2s-complement) 
     * ASN.1 INTEGER value.                                                 <P>
     * 
     * The 'Integer' in this method's title refers to the ASN.1 INTEGER type,
     * not a Java <i>int</i> or <i>Integer</i>.  An ASN.1 INTEGER is a signed 
     * (2s-complement) binary integer of unspecified range.  This function 
     * actually returns a variable of type <i>long</i>.
     * 
     * @param data The array of bytes to be parsed.                         <P>
     *
     * @return A <i>long</i> containing the parsed number.                  <P>
     *
     */
    public static long parseInteger( byte[] data )
    {
        StringBuffer bitBuffer;
        int i;
        long value;

        if ( data.length > 8 ) 
            return ( -1 );      // really should throw an exception instead
        else {
            bitBuffer = new StringBuffer();
            for ( i = 0; i < data.length; i++ ) {
                String thisByte = Integer.toHexString( (int)data[ i ] );
                if ( thisByte.length() > 2 )
                    // Java annoyingly pads negative bytes with leading F's
                    bitBuffer.append( thisByte.substring( thisByte.length()-1 ));
                else bitBuffer.append( thisByte );
            } 
            value = Long.parseLong( bitBuffer.toString(), 16 );
            return ( value );
        }    
    } // ----------------------------------------------------------------------


    // --------------------------------------------------- parseUnsignedInteger
    /**
     * Translates an array of bytes into a single unsigned ASN.1 INTEGER 
     * value.                                                               <P>
     *
     * Like the <i>parseInteger()</i> method, this method returns a variable of
     * type <i>long</i>.  Unlike <i>parseInteger()</i>, the number is converted
     * into an unsigned value.
     *    
     * @param bits The array of bytes to be parsed.                         <P>
     *
     * @return A <i>long</i> containing the parsed number.                  <P>
     *
     */
    public static long parseUnsignedInteger( byte[] bits )
    {
        String byteString;
        long value;

        if ( bits.length > 8 ) 
            return ( -1 );      // really should throw an exception instead
        else {
            byteString = getByteString( bits );        
            value = Long.parseLong( byteString, 16 );
            if ( value < 0 )
                value += (long)Math.pow( 16, byteString.length() - 1 );
            return ( value );
        }    
    } // ----------------------------------------------------------------------


    // ------------------------------------------------------- parseTextOrBytes
    /** 
     * This procedure attempts to 'guess' whether a byte sequence should be
     * displayed as ASCII text or as a hexadecimal byte string.  If the 
     * byte array consists entirely of printable ASCII characters (with or
     * without a terminating null character), then it is translated into an
     * ASCII string.  Otherwise, it is returned in formatted hexadecimal
     * notation.  (Single-byte arrays are always returned as hex numbers.)  <P>
     *
     * @param bytes The array of bytes to be parsed.                        <P>
     *
     * @return The decoded <i>String</i>.                                   <P>
     *
     */
    public static String parseTextOrBytes( byte[] bytes )
    {
        String value;
        int i,
            ibyte;
        boolean isText = true;

        // one-byte values are almost certainly not text...        
        if ( bytes.length == 1 ) isText = false;
        else for ( i = 0; i < bytes.length; i++ ) {
            if (( ibyte = (int)bytes[i] ) < 0 ) ibyte += 256; 
            // check each byte
            if (( ibyte > 0 ) && ( ibyte < 32 ))               isText = false;
            else if ( ibyte > 127 )                            isText = false;
            else if (( ibyte == 0 ) && ( i+1 < bytes.length )) isText = false; 

            if ( ! isText) break;
        }

        if ( isText ) value = ("\"" + new String( bytes ) + "\"");
        else          value = "0x" + getByteString( bytes );

        return ( value );
    } // ----------------------------------------------------------------------

    
    // ---------------------------------------------------------- getByteString
    /**
     * Translates an array of bytes into a formatted hexadecimal string.    <P>
     *
     * @param bytes The array of bytes to be parsed.                        <P>
     *
     * @return The formatted <i>String</i>.                                 <P>
     *
     */
    public static String getByteString( byte[] bytes )
    {
        int ibyte,
            i;
        StringBuffer dataBuf = new StringBuffer();

        for ( i = 0; i < bytes.length; i++ ) {
            if (( ibyte = (int)bytes[i] ) < 0 ) ibyte += 256; // Convert byte to unsigned value
            if ( ibyte < 16 ) dataBuf.append("0" + Integer.toHexString( ibyte ));
            else dataBuf.append( Integer.toHexString( ibyte ));
        } 
        return ( dataBuf.toString() );
    } // ----------------------------------------------------------------------

} // ==========================================================================




// ============================================================================
/**
 * A structure representing an ASN.1 encoded element.
 *
 */
class EncodedElement
{

               /** 
                * The value of the identifier byte(s) as a single integer. 
                */
    public int identifier;
    
    public int idClass;
    public int idEncoding;
    public int idTag;


               /** 
                * The length, in bytes, of the data portion of this element.
                * This is the same as the length of the data[] array, below.
                */
    public int length;        

                /** 
                 * The number of bytes in the element before the start of the
                 * data bytes.  This is equal to the number of identifier bytes 
                 * plus the number of length bytes.   
                 */
    public int offsetinc;

                /** 
                 * An array of bytes containing the element data.      
                 */
    public byte data[];

} // ==========================================================================




// ============================================================================
/**
 * Class representing an SMI VariableBinding type.
 *
 */
class VariableBinding
{
                /**
                 * The variable name (an OBJECT IDENTIFIER) in numeric form.
                 */
    public String numericOID;
    
                /**
                 * The variable name in translated form (as much of it as can
                 * be determined).
                 */
    public String textualOID;
    
                /**
                 * The variable's value as a hexadecimal byte string.
                 */
    public String value;
    
                /**
                 * The variable's value, translated to ASCII text if it 
                 * contains only printable characters; a hex string otherwise.
                 */
    public String textValue;

    // ------------------------------------------------------------ constructor
    VariableBinding( byte[] nameBytes, byte[] valueBytes ) {
        numericOID = SMIHandler.parseOID( nameBytes );
        textualOID = SMIHandler.translateOID( numericOID );
        value      = SMIHandler.getByteString( valueBytes );
        textValue  = SMIHandler.parseTextOrBytes( valueBytes );
    } // ----------------------------------------------------------------------
    
} // ==========================================================================

