package trapmon;

import java.util.*;
import java.net.*;

/**
 * Encapsulates data from a single SNMP message.
 *
 * @author Alex Taylor, copyright (C) 2003
 *
 */
public class SnmpMessage extends Object
{

    // Some ASN.1 element types specific to SNMP messages...
                /**
                 * The SNMP GetRequest-PDU message type.
                 */
    public static final int ASN_SNMP_GETREQUEST     = 0xA0;

                /**
                 * The SNMP GetNextRequest-PDU message type.
                 */
    public static final int ASN_SNMP_GETNEXTREQUEST = 0xA1;

                /**
                 * The SNMP GetResponse-PDU message type.
                 */
    public static final int ASN_SNMP_GETRESPONSE    = 0xA2;

                /**
                 * The SNMP SetRequest-PDU message type.
                 */
    public static final int ASN_SNMP_SETREQUEST     = 0xA3;

                /**
                 * The SNMP Trap-PDU message type.
                 */
    public static final int ASN_SNMP_TRAP           = 0xA4;


    public InetAddress originator;

    public int     type;

    public boolean isValid;

    public long    requestID;

    public int     version,
                   errorStatus,
                   errorIndex,
                   genericTrapNumber,
                   specificTrapNumber;

    public long    timeTicks;

    public String  community,
                   messageName,
                   errorStatusName,
                   numericOID,
                   textualOID,
                   genericTrapName,
                   specificTrapName,
                   agentAddress,
                   timeElapsed;

    public Vector  variables;


    // ------------------------------------------------------------ constructor
    /**
     * Creates and initializes the <i>SnmpMessage</i> object.               <P>
     *
     * A SNMP message is an ASN.1 element of the form                    <PRE>
     *   SEQUENCE { version, community, data }                           </PRE>                           
     * encoded using SMI.  In general, the application will receive a raw 
     * UDP packet containing the SNMP message, use several calls to 
     * <i>SMIHandler.decodeElement()</i> to parse the SNMP version number, 
     * community name, and message data, and then pass that information to 
     * <tt>new SnmpMessage()</tt> via this constructor.
     *
     * @param versionElement An <i>EncodedElement</i> containing the SNMP
     *   version number.
     *
     * @param communityElement An <i>EncodedElement</i> containing the SNMP
     *   community name.
     *
     * @param dataElement An <i>EncodedElement</i> containing the SNMP 
     *   message data.
     *
     * @param sender An <i>InetAddress</i> containing the IP address
     *   from whence the SNMP message has come.
     *
     */
    public SnmpMessage( EncodedElement versionElement,
                        EncodedElement communityElement,
                        EncodedElement dataElement,
                        InetAddress    sender            ) {

        isValid = false;
        
        originator = sender;
        version    = (int) SMIHandler.parseUnsignedInteger( versionElement.data ) + 1;
        community  = new String( communityElement.data );

        type = dataElement.identifier;
        switch ( dataElement.identifier ) {
            case ASN_SNMP_GETREQUEST:     parseMessagePDU( dataElement.data ); break;
            case ASN_SNMP_GETNEXTREQUEST: parseMessagePDU( dataElement.data ); break;
            case ASN_SNMP_GETRESPONSE:    parseMessagePDU( dataElement.data ); break;
            case ASN_SNMP_SETREQUEST:     parseMessagePDU( dataElement.data ); break;
            case ASN_SNMP_TRAP:           parseTrapPDU( dataElement.data );    break;
            default:                      type = 0;                            break;
        }    

    } // ----------------------------------------------------------------------


    // ----------------------------------------------------------- parseTrapPDU
    /**
     * Decodes the data field of a SNMP Trap-PDU.
     *                                                                      <P>
     * A SNMP Trap-PDU takes the format:
     *                                                                    <PRE>
     * IMPLICIT SEQUENCE {
     *     enterprise           OBJECT IDENTIFIER       ,
     *     agent-addr           NetworkAddress          ,
     *     generic-trap         INTEGER {0,1,2,3,4,5,6} ,
     *     specific-trap        INTEGER                 ,
     *     time-stamp           TimeTicks               ,
     *     variable-bindings    VarBindList
     * }                                                                 </PRE>
     *
     */
    protected void parseTrapPDU( byte[] trap )
    {
        EncodedElement enterpriseElement,
                       agentAddrElement,
                       genericTrapElement,
                       specificTrapElement,
                       timeStampElement,
                       variableBindings;
        int offset,
            trapNumber;

        offset = 0;

        // Enterprise OID
        enterpriseElement = SMIHandler.decodeElement( trap, offset );
        numericOID        = SMIHandler.parseOID( enterpriseElement.data );
        textualOID        = SMIHandler.translateOID( numericOID );
        offset += ( enterpriseElement.offsetinc + enterpriseElement.length );

        // Agent's IP address        
        agentAddrElement = SMIHandler.decodeElement( trap, offset );
        agentAddress     = SMIHandler.parseNetworkAddress( agentAddrElement.data );
        offset += (agentAddrElement.offsetinc + agentAddrElement.length);

        // Generic trap ID
        genericTrapElement = SMIHandler.decodeElement( trap, offset );
        genericTrapNumber  = (int) SMIHandler.parseInteger( genericTrapElement.data );
        genericTrapName    = getGenericTrapName( genericTrapNumber );
        offset += (genericTrapElement.offsetinc + genericTrapElement.length);

        // Enterprise-specific trap ID
        specificTrapElement = SMIHandler.decodeElement( trap, offset );
        specificTrapNumber  = (int) SMIHandler.parseInteger( specificTrapElement.data );
        offset += (specificTrapElement.offsetinc + specificTrapElement.length);

        // Timestamp
        timeStampElement = SMIHandler.decodeElement( trap, offset );
        if ( timeStampElement.identifier == SMIHandler.ASN_SMI_TIMETICKS ) {
            timeElapsed  = SMIHandler.getByteString( timeStampElement.data );
            timeTicks    = SMIHandler.parseUnsignedInteger( timeStampElement.data );
        } else {
            timeElapsed  = "unknown";
            timeTicks    = 0;
        }            
        offset += (timeStampElement.offsetinc + timeStampElement.length);

        // List of trap-specific variables
        variables = new Vector();        
        if ( offset < ( trap.length - 1 )) {
            variableBindings = SMIHandler.decodeElement( trap, offset );
            offset += (variableBindings.offsetinc + variableBindings.length);

            if (( variableBindings.identifier == SMIHandler.ASN_SEQUENCE ) && ( variableBindings.length > 0 )) {
                boolean endOfData = false;
                EncodedElement variable;
                VariableBinding parsedVariable;
                offset = 0;
                while ( ! endOfData ) {
                    variable = SMIHandler.decodeElement( variableBindings.data, offset );
                    parsedVariable = parseMibVariable( variable );
                    if ( parsedVariable != null )
                        variables.addElement( parsedVariable );
                    offset += (variable.offsetinc + variable.length);
                    if ( offset >= variableBindings.length ) endOfData = true;
                }
            }
        }
        isValid = true;
        
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------- getGenericTrapName
    /**
     * Translate a generic trap number into the corresponding trap name.
     *
     */
    protected String getGenericTrapName( int trapNumber )
    {
        String trapMessage;
        switch ( trapNumber ) {
            case  0: trapMessage = "coldStart";             break;
            case  1: trapMessage = "warmStart";             break;
            case  2: trapMessage = "linkDown";              break;
            case  3: trapMessage = "linkUp";                break;
            case  4: trapMessage = "authenticationFailure"; break;
            case  5: trapMessage = "egpNeighbourLoss";      break;
            case  6: trapMessage = "enterpriseSpecific";    break;
            default: trapMessage = "unknown trap code";     break;
        }
        return ( trapMessage );
    } // ----------------------------------------------------------------------


    // -------------------------------------------------------- parseMessagePDU
    /**
     * Decodes the data field of any of the non-trap SNMP messages.
     *                                                                      <P>
     * The non-trap SNMP message types (GetRequest-PDU, GetNextRequest-PDU, 
     * GetResponse-PDU and SetRequest-PDU) all take the format:
     *                                                                    <PRE>
     * IMPLICIT SEQUENCE {
     *     request-id           RequestID   ,
     *     error-status         ErrorStatus ,
     *     error-index          ErrorIndex  ,
     *     variable-bindings    VarBindList
     * }                                                                 </PRE>
     *
     */
    protected void parseMessagePDU( byte[] pdu )
    {
        EncodedElement requestIdElement,
                       errorStatusElement,
                       errorIndexElement,
                       variableBindings;
        int offset;

        offset = 0;

        // The name of the SNMP message type
        messageName = getMessageName( type );

        // Request ID
        requestIdElement = SMIHandler.decodeElement( pdu, offset );
        requestID        = SMIHandler.parseInteger( requestIdElement.data );
        offset += (requestIdElement.offsetinc + requestIdElement.length);

        // Error Status
        errorStatusElement = SMIHandler.decodeElement( pdu, offset );
        errorStatus        = (int) SMIHandler.parseInteger( errorStatusElement.data );
        errorStatusName    = getErrorStatusName( errorStatus );
        offset += (errorStatusElement.offsetinc + errorStatusElement.length);

        // Error Index
        errorIndexElement = SMIHandler.decodeElement( pdu, offset );
        errorIndex        = (int) SMIHandler.parseInteger( errorIndexElement.data );
        offset += (errorIndexElement.offsetinc + errorIndexElement.length);

        // List of message-specific variables
        variables = new Vector();        
        if ( offset < pdu.length ) {
            variableBindings = SMIHandler.decodeElement( pdu, offset );
            offset += (variableBindings.offsetinc + variableBindings.length);

            if ( variableBindings.identifier == SMIHandler.ASN_SEQUENCE ) {
                boolean endOfData = false;
                EncodedElement variable;
                VariableBinding parsedVariable;
                offset = 0;
                while ( ! endOfData ) {
                    variable = SMIHandler.decodeElement( variableBindings.data, offset );
                    parsedVariable = parseMibVariable( variable );
                    if ( parsedVariable != null )
                        variables.addElement( parsedVariable );
                    offset += (variable.offsetinc + variable.length);
                    if ( offset >= variableBindings.length ) endOfData = true;
                }
            }
        }
        isValid = true;
        
    } // ----------------------------------------------------------------------


    // --------------------------------------------------------- getMessageName
    /**
     * Translate a message number into the corresponding message name.
     *
     */
    protected String getMessageName( int id )
    {
        String message;
        switch ( id ) {
            case ASN_SNMP_GETREQUEST:     message = new String("GetRequest");     break;
            case ASN_SNMP_GETNEXTREQUEST: message = new String("GetNextRequest"); break;
            case ASN_SNMP_GETRESPONSE:    message = new String("GetResponse");    break;
            case ASN_SNMP_SETREQUEST:     message = new String("SetRequest");     break;
            default:                      message = "Unknown";                    break;
        }
        return ( message );
    } // ----------------------------------------------------------------------


    // ----------------------------------------------------- getErrorStatusName
    /**
     * Translate a SNMP error number into the corresponding error name.
     *
     */
    protected String getErrorStatusName( int error )
    {
        String errorMessage;
        switch ( error ) {
            case  0: errorMessage = "noError";              break;
            case  1: errorMessage = "tooBig";               break;
            case  2: errorMessage = "noSuchName";           break;
            case  3: errorMessage = "badValue";             break;
            case  4: errorMessage = "readOnly";             break;
            case  5: errorMessage = "genErr";               break;
            default: errorMessage = "unknown error status"; break;
        }
        return ( errorMessage );
    } // ----------------------------------------------------------------------


    // ---------------------------------------------------------- parseSequence
    /**
     * Parses a MIB variable binding.
     *
     */
    VariableBinding parseMibVariable( EncodedElement varBindElement )
    {
        VariableBinding variable;
        EncodedElement name, 
                       value;
        int offset = 0;

        if ( varBindElement.identifier != SMIHandler.ASN_SEQUENCE ) 
            return ( null );
        else {
            name = SMIHandler.decodeElement( varBindElement.data, offset );
            offset += (name.offsetinc + name.length);
            
            value = SMIHandler.decodeElement( varBindElement.data, offset );
            offset += (value.offsetinc + value.length);

            variable = new VariableBinding( name.data, value.data );
            return ( variable );
        }

    } // ----------------------------------------------------------------------
    

} // ==========================================================================

